#include <eopcc/utility.hpp>
#include <eopcc/utm_zone.hpp>
#include <eopcc/cloud_io.hpp>
#include <tinyxml/tinyxml2.h>

#include <sstream>
#include <memory>

#ifdef GDAL_SUPPORT
// GDAL includes
#include <gdal_priv.h>
#include <ogr_spatialref.h>
#endif

namespace eopcc {

#ifdef GDAL_SUPPORT

/// Validates a cloud file one point at a time
bool ValidateCloud(const std::string& cloud_file, const std::string& kml_file) {
	std::unique_ptr<PointCloudReader> reader(CreateReader(cloud_file));

    PolygonExtent extent;
    bool extent_valid = !kml_file.empty();

    if (extent_valid) {
        try {
            extent = ExtentFromKML(kml_file);
        } catch (std::runtime_error& e) {
            extent_valid = false;
        }
    }

    size_t num_points = reader->NumRemainingPoints();

    std::string wkt_str = reader->GetSeedCloud().wkt_spatial_ref;
    if (wkt_str.empty()) {
	    std::cout << "File opened, no spatial reference found" << std::endl;
	    std::cout << "Assuming file is in UTM coordinates in the same zone as the KML" << std::endl;
    } else {
        std::cout << "Found spatial reference: " << wkt_str << std::endl;
    }

    if (extent_valid) {
	    if (wkt_str.empty()) {
		    extent.MatchCoordinatesUTM();
	    } else {
		    extent.MatchCoordinates(wkt_str);
	    }
    }

    if (num_points < 1) {
      throw ValidationError("File opened, but no points found");
    }

    int inbounds_count = 0;
    for (size_t index=0; index < num_points; ++index) {
	    PointCloud cloud = reader->ReadSubCloud(1);

	    if (cloud.size() == 0 || cloud.size() > 1) {
		    throw ValidationError("Unexcepted end to cloud");
	    }


	    if (extent_valid) {
		    if (extent.contains(cloud.loc[0])) {
			    inbounds_count++;
		    }
	    }
    }

    if (extent_valid) {
	    if (inbounds_count == 0) {
		    throw ValidationError("No points inside KML bounds");
	    }

	    std::cout << "Tested bounds: " << ((double)inbounds_count) / num_points * 100 << "% of points inside KML bounds" << std::endl;
    }

    return true;
}


void PolygonExtent::MatchCoordinatesUTM() {
	assert(sides.size() > 0);

	int zone = GetUTMZone(sides[0].start.y, sides[0].start.x);
	bool north = IsNorthernHemisphere(sides[0].start.y);
	MatchCoordinatesUTM(zone, north);
}

void PolygonExtent::MatchCoordinatesUTM(int utm_zone, bool north) {
	OGRSpatialReference utm_srs;
	utm_srs.SetWellKnownGeogCS("WGS84");
	utm_srs.SetUTM(utm_zone, north);

	char* wkt = NULL;
	utm_srs.exportToWkt(&wkt);
	MatchCoordinates(wkt);
	OGRFree(wkt);
}

void PolygonExtent::MatchCoordinates(const std::string& wkt_string) {
  OGRSpatialReference utm_srs(wkt_string.c_str());
  OGRSpatialReference latlon_srs;
  latlon_srs.CopyGeogCSFrom(&utm_srs);

  OGRCoordinateTransformation* coordinate_transform;

  coordinate_transform = OGRCreateCoordinateTransformation(&latlon_srs, &utm_srs);

  for (eopcc::Line& line : sides) {
    coordinate_transform->Transform(1, &line.start.x, &line.start.y);
    coordinate_transform->Transform(1, &line.end.x, &line.end.y);
  }
}
#endif

PolygonExtent ExtentFromKML(const std::string& kml_file) {
	using namespace tinyxml2;

	XMLDocument doc;
	XMLError err = doc.LoadFile(kml_file.c_str());

    if (err) {
        throw std::runtime_error(kml_file + " doesn't exist");
    }

	XMLElement* root_element = doc.RootElement();


	const char* nodeName = root_element->Value();
	XMLNode* document =  root_element->FirstChildElement("Document");
	if (!document) {
		throw std::runtime_error("Failed to parse document");
	}

	XMLNode *placemark =  document->LastChildElement("Placemark");
	if (!placemark)
	{
		throw std::runtime_error("Failed to parse placemark");
	}

	XMLNode* polygon =  placemark->LastChildElement("Polygon");
	if (!polygon)
	{
		throw std::runtime_error("Failed to parse polygon");
	}

	XMLNode *outer_boundary_is =  polygon->FirstChildElement("outerBoundaryIs");
	if (!outer_boundary_is)
	{
		throw std::runtime_error("Failed to parse outerBoundaryIs");
	}

	XMLNode* linear_ring =  outer_boundary_is->FirstChildElement("LinearRing");
	if (!linear_ring)
	{
		throw std::runtime_error("Failed to parse LinearRing");
	}

	XMLNode *coords =  linear_ring->FirstChild ();
	if (!coords)
	{
		throw std::runtime_error("Failed to parse coords");
	}

	XMLNode *xml_contents = coords->FirstChild();
	if (!xml_contents) {
		throw std::runtime_error("Failed to parse coords contents");
	}

	std::string contents = xml_contents->Value();


	std::vector<Point> points;

	std::stringstream ss;
	ss << contents;

	for (;;) {
		std::string value;
		Point pt;

		try {
			std::getline(ss, value, ',');
			pt.x = std::stod(value);
			std::getline(ss, value, ',');
			pt.y = std::stod(value);
			std::getline(ss, value, '\n');
			pt.z = std::stod(value);
		} catch (std::invalid_argument& e) {
			break;
		}

		points.push_back(pt);
	}

	PolygonExtent poly(points);
	return poly;
}
// Checks for intersection between a ray along the x axis starting at
// the point and line.
bool ray_intersects_segment(Point point, const Line& line) {
	const double epsilon = std::numeric_limits<double>::min();
	Point lower = line.start;
	Point upper = line.end;

	if (point.y == lower.y || point.y == upper.y) {
		point.y += epsilon;
	}

	if (point.y < lower.y || point.y > upper.y) {
		return false;
	} else if (point.x > std::max(lower.x, upper.x)) {
		return false;
	} else {
		if (point.x < std::min(lower.x, upper.x)) {
			return true;
		} else {
			double m_line;
			double m_ray;
			if (lower.x != upper.x) {
				m_line = (upper.y - lower.y)/(upper.x - lower.x);
			} else {
				m_line = std::numeric_limits<double>::max();
			}
			if (lower.x != point.x) {
				m_ray = (point.y - lower.y)/(point.x - lower.x);
			} else {
				m_ray = std::numeric_limits<double>::max();
			}
			if (m_ray >= m_line) {
				return true;
			} else {
				return false;
			}
		}
	}
}

bool PolygonExtent::contains(const Point& point) const {
	int intersection_count = 0;

	for (const Line& line : sides) {
		if (ray_intersects_segment(point, line)) {
			intersection_count++;
		}
	}

	// If the intesection count is odd, the point is inside the polygon
	return (intersection_count % 2 == 1);
}


}
